#!/usr/bin/env python
"""test program to exercise multicast send/receive."""

from __future__ import print_function


import sys, time
from socket import socket, AF_INET, AF_INET6, SOCK_DGRAM, IPPROTO_UDP,          \
    SOL_SOCKET, SO_REUSEADDR, IPPROTO_IP, IPPROTO_IPV6, IP_ADD_MEMBERSHIP,          \
    IPV6_JOIN_GROUP, IP_MULTICAST_TTL, IPV6_MULTICAST_HOPS, IP_MULTICAST_LOOP,          \
    IPV6_MULTICAST_LOOP, inet_pton, timeout as TimeoutError
import socket as net
from struct import pack

from golang import b, u, go, func, defer, chan
from golang import os as gos, sync, context, syscall
from golang.os import signal


group4 = '224.1.1.1'
group6 = 'ff1e::1'
port   =  5678

def _udpsockfor(ipaddr): # -> (sk, isipv6)
    ip6 = (':' in ipaddr)  # autodetect ip4/ip6 based on group addr
    if ip6:
        sk = socket(AF_INET6, SOCK_DGRAM, IPPROTO_UDP)
    else:
        sk = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)
    return (sk, ip6)


# mjoin_tx returns socket prepated to send to multicast group:port.
def mjoin_tx(group, port, ttl=100):
    sk, ip6 = _udpsockfor(group)

    # set ttl and default destination address (for tx)
    if ip6:
        sk.setsockopt(IPPROTO_IPV6, IPV6_MULTICAST_HOPS, ttl)
    else:
        sk.setsockopt(IPPROTO_IP, IP_MULTICAST_TTL, ttl)

    # NOTE do _not_ connect. If connected, Linux may permanently associate destination with wrong route.
    # -> use sk.sendto() instead of sk.send() in txloop.
    #sk.connect((group, port))

    # turn off automatic loopback.
    # We _do_ want to see what we send looping back to ourselves.
    # We _do_ receive it back because we configure default multicast route to go to lo.
    # When IP_MULTICAST_LOOP=1 the kernel has also dedicated path to send
    # packets back irregardless of what is the route. Turn that later path off
    # to avoid receiving what we send twice.
    if ip6:
        sk.setsockopt(IPPROTO_IPV6, IPV6_MULTICAST_LOOP, 0)
    else:
        sk.setsockopt(IPPROTO_IP, IP_MULTICAST_LOOP, 0)

    return sk

# mjoin_rx returns socket prepated to receive from multicast group:port.
def mjoin_rx(group, port):
    sk, ip6 = _udpsockfor(group)

    sk.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
    sk.bind((group, port))  # works for reception even if we would want to rx from multiple ifaces

    # join the group on mcast-rx (preferrably, if it exists), or loopback.
    #
    # mcast-rx is used instead of loopback for receiving multicast, because
    # with IPv6 Linux rejects fowarding multicast traffic to lo.
    # See https://lab.nexedi.com/kirr/linux/commit/25bcc76020c5 for details.
    #
    # NOTE we used to join the group on all interfaces, but this results in
    # multiple delivery of the same packet, e.g. if local node forwards that
    # packet from its external interface A to B.
    mcast_ifv = ("mcast-rx", "lo")  # XXX change lo to interface corresponding to daddr=group route?
    ifidx = None
    for ifname in mcast_ifv:
        try:
            ifidx = net.if_nametoindex(ifname)
        except OSError:
            continue
        break
    if ifidx is None:
        raise RuntimeError("mjoin_rx: cannot find suitable interface; tried: %s" % (mcast_ifv,))


    print("joining %s @ [%d]%s" % (group, ifidx, ifname))
    bifidx = pack("@i", ifidx)
    if ip6:
        mreq = inet_pton(AF_INET6, group) + \
               bifidx
        sk.setsockopt(IPPROTO_IPV6, IPV6_JOIN_GROUP, mreq)  # = IPV6_ADD_MEMBERSHIP
    else:
        mreq = inet_pton(AF_INET, group)     + \
               inet_pton(AF_INET, '0.0.0.0') + \
               bifidx
        sk.setsockopt(IPPROTO_IP, IP_ADD_MEMBERSHIP, mreq)

    return sk


def txloop(ctx, sk, daddr, pkt):
    i = 0
    while 1:
        if ctx.err():
            raise ctx.err()
        i += 1
        pkt_ = pkt + (b'.%d' % i)
        print("tx: %s ..." % u(pkt_))
        sk.sendto(pkt_, daddr)
        time.sleep(1)

def rxloop(ctx, sk):
    sk.settimeout(1)
    while 1:
        if ctx.err():
            raise ctx.err()
        try:
            pkt = sk.recv(4096)
        except TimeoutError:
            continue
        print("rx: %s" % u(pkt))


@func
def main():
    action = sys.argv[1]

    G4 = (group4, port)
    G6 = (group6, port)
    G = None
    if "4" in action:
        G = G4
    if "6" in action:
        G = G6
    if G is None:
        raise RuntimeError("4/6 must be specified")

    # prepare ctx that is canceled on SIGINT/SIGTERM
    ctx, cancel = context.with_cancel(context.background())
    sigq = chan(1, dtype=gos.Signal)
    signal.Notify(sigq, syscall.SIGINT, syscall.SIGTERM)
    def _():
        signal.Stop(sigq)
        sigq.close()
    defer(_)
    def _():
        sig, ok = sigq.recv_()
        if not ok:
            return
        print("# %s" % sig)
        cancel()
    go(_)
    defer(cancel)

    wg = sync.WorkGroup(ctx)
    if "tx" in action:
        sktx = mjoin_tx(*G)
        wg.go(txloop, sktx, G, b(sys.argv[2]))
    if "rx" in action:
        skrx = mjoin_rx(*G)
        wg.go(rxloop, skrx)

    wg.wait()

if __name__ == '__main__':
    main()
